@@ -56,42 +56,42 @@ import org.jboss.util.collection.Iterators;
 
 /**
  * BaseClassLoader.
- * 
+ *
  * @author <a href="adrian@jboss.com">Adrian Brock</a>
  * @version $Revision: 1.1 $
  */
-public class BaseClassLoader extends SecureClassLoader implements BaseClassLoaderMBean, RealClassLoader 
+public class BaseClassLoader extends SecureClassLoader implements BaseClassLoaderMBean, RealClassLoader
 {
    /** The log */
    private static final Logger log = Logger.getLogger(BaseClassLoader.class);
 
    /** The lock object */
    private ReentrantLock lock = new ReentrantLock(true);
-   
+
    /** The policy for this classloader */
    private ClassLoaderPolicy policy;
-   
+
    /** Our Loader front end */
    private DelegateLoader loader;
-   
+
    /** The loaded classes */
    private Map<String, String> loadedClasses = new ConcurrentHashMap<String, String>();
-   
+
    /** Our resource cache */
    private Map<String, URL> resourceCache;
-   
+
    /** Our black list */
    private Map<String, String> blackList;
 
    /** A cached version of toLongString() */
    private String cachedToString;
-   
+
    /**
     * Create a new ClassLoader with no parent.
-    * 
+    *
     * @param policy the policy
-    * @throws IllegalArgumentException for a null policy 
-    * @throws IllegalStateException if the policy is already associated with a classloader 
+    * @throws IllegalArgumentException for a null policy
+    * @throws IllegalStateException if the policy is already associated with a classloader
     */
    public BaseClassLoader(ClassLoaderPolicy policy)
    {
@@ -110,7 +110,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
 
       if (basePolicy.isBlackListable())
          blackList = new ConcurrentHashMap<String, String>();
-      
+
       log.debug("Created " + this + " with policy " + policy.toLongString());
    }
 
@@ -118,7 +118,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
    {
       return policy.getObjectName();
    }
-   
+
    public ObjectName getClassLoaderDomain()
    {
       BaseClassLoaderPolicy basePolicy = policy;
@@ -130,7 +130,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
    {
       return policy.getName();
    }
-   
+
    public boolean isBlackListable()
    {
       BaseClassLoaderPolicy basePolicy = policy;
@@ -175,7 +175,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
       }
       return result;
    }
-   
+
    public String listPolicyDetails()
    {
       return policy.toLongString();
@@ -186,18 +186,18 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
       final Class<?> clazz = loadClass(name);
       if (clazz == null)
          return null;
-      
+
       ClassLoader cl = AccessController.doPrivileged(new PrivilegedAction<ClassLoader>()
       {
          public ClassLoader run()
          {
-            return clazz.getClassLoader(); 
+            return clazz.getClassLoader();
          }
       });
-      
+
       if (cl != null && cl instanceof RealClassLoader)
          return ((RealClassLoader) cl).getObjectName();
-      
+
       return null;
    }
 
@@ -222,7 +222,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
 
    /**
     * Get the policy.
-    * 
+    *
     * @return the policy.
     */
    ClassLoaderPolicy getPolicy()
@@ -232,7 +232,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
 
    /**
     * Get the loader.
-    * 
+    *
     * @return the loader.
     */
    DelegateLoader getLoader()
@@ -249,12 +249,12 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
 
       if (name == null)
          return null;
-      
+
       // Did we already load this package?
       Package result = super.getPackage(name);
       if (result != null && trace)
          log.trace(this + " already loaded package " + name + " " + result.getName());
-      
+
       // Not already loaded use the domain
       if (result == null)
       {
@@ -266,14 +266,14 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
             log.trace(this + " getPackage " + name + " domain=" + domain);
          result = domain.getPackage(this, name);
       }
-      
+
       // Still not found
       if (result == null)
       {
          if (trace)
             log.trace(this + " package not found " + name);
       }
-      
+
       return result;
    }
 
@@ -295,7 +295,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
 
    /**
     * Get a package locally
-    * 
+    *
     * @param name the package name
     * @return the package
     */
@@ -306,7 +306,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
 
    /**
     * Get the packages locally
-    * 
+    *
     * @param packages the packages
     */
    void getPackagesLocally(Set<Package> packages)
@@ -315,10 +315,10 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
       for (Package pkg : pkgs)
          packages.add(pkg);
    }
-   
+
    /**
     * Check to see if the class is already loaded
-    * 
+    *
     * @param name the name of the class
     * @param trace whether trace is enabled
     * @return the class is if it is already loaded, null otherwise
@@ -347,9 +347,31 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
       return result;
    }
 
+   /**
+    * Check the cache and blacklist
+    *
+    * @param name the name of the class
+    * @param failIfBlackListed <code>true</code> if a blacklisted class should
+    *                          result in ClassNotFoundException; <code>false</code>
+    *                          if a <code>null</code> return value is acceptable
+    * @param trace whether trace is enabled
+    * @return the class is if it is already loaded, null otherwise
+    * @throws ClassNotFoundException when the class is blacklisted and 
+    *                               <code>failIfBlackListed</code> is <code>true</code>
+    */
+   protected Class<?> checkCacheAndBlackList(String name, boolean failIfBlackListed, boolean trace) throws ClassNotFoundException
+   {
+      BaseClassLoaderPolicy basePolicy = policy;
+      BaseClassLoaderDomain domain = basePolicy.getClassLoaderDomain();
+      if (domain == null)
+         return null;
+
+      return domain.checkClassCacheAndBlackList(this, name, null, basePolicy.isImportAll(), false);
+   }
+
    /**
     * Find the classloader for a class but don't load the class
-    * 
+    *
     * @param className the class name
     * @return the classloader
     * @throws ClassNotFoundException if the class is not found
@@ -369,11 +391,11 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
 
       ClassLoaderUtils.checkClassName(className);
       String path = ClassLoaderUtils.classNameToPath(className);
-      
+
       Loader loader = domain.findLoader(this, path, basePolicy.isImportAll(), true);
       if (loader == null)
          throw new ClassNotFoundException("Class " + className + " not found from " + this);
-      
+
       // This is a bit ugly but we can't abstract this behind an interface because
       // that would make the methods public
       if (loader instanceof BaseClassLoaderSource)
@@ -389,10 +411,10 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
       boolean trace = log.isTraceEnabled();
       if (trace)
          log.trace(this + " loadClass " + name + " resolve=" + resolve);
-      
+
       // Validate the class name makes sense
       ClassLoaderUtils.checkClassName(name);
-      
+
       // Did we already load this class?
       Class<?> result = isLoadedClass(name, trace);
 
@@ -407,13 +429,17 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
       }
       if (result != null)
          return result;
-      
+
+      result = checkCacheAndBlackList(name, false, trace);
+      if (result != null)
+         return result;
+
       synchronized (this)
       {
          // Not already loaded use the domain
          if (result == null)
             result = loadClassFromDomain(name, trace);
-         
+
          // Still not found
          if (result == null)
          {
@@ -421,7 +447,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
                log.trace(this + " class not found " + name);
             throw new ClassNotFoundException(name + " from " + toLongString());
          }
-         
+
          // Link the class if requested
          if (resolve)
          {
@@ -429,7 +455,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
                log.trace(this + " resolveClass " + ClassLoaderUtils.classToString(result));
             resolveClass(result);
          }
-         
+
          return result;
       }
    }
@@ -469,10 +495,10 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
          domain.getResources(this, name, resourceURLs);
       return resourceURLs;
    }
-   
+
    /**
     * Try to load the class locally
-    * 
+    *
     * @param name the class name
     * @return the class
     */
@@ -483,7 +509,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
 
    /**
     * Try to load the class locally
-    * 
+    *
     * @param name the class name
     * @param trace whether trace is enabled
     * @return the class if found
@@ -500,7 +526,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
 
       // Look for the resource
       final String resourcePath = ClassLoaderUtils.classNameToPath(name);
-      
+
       result = AccessController.doPrivileged(new PrivilegedAction<Class<?>>()
       {
          public Class<?> run()
@@ -515,7 +541,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
 
             // Load the bytecode
             byte[] byteCode = ClassLoaderUtils.loadByteCode(name, is);
-            
+
             // Let the policy do things before we define the class
             BaseClassLoaderPolicy basePolicy = policy;
             ProtectionDomain protectionDomain = basePolicy.getProtectionDomain(name, resourcePath);
@@ -529,7 +555,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
             {
                throw new RuntimeException("Unexpected error transforming class " + name, t);
             }
-            
+
             // Create the package if necessary
             URL codeSourceURL = null;
             if (protectionDomain != null)
@@ -539,7 +565,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
                   codeSourceURL = codeSource.getLocation();
             }
             definePackage(name, codeSourceURL);
-            
+
             // Finally we can define the class
             Class<?> result;
             if (protectionDomain != null)
@@ -551,15 +577,15 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
             return result;
          }
       }, policy.getAccessControlContext());
-      
+
       loadedClasses.put(name, name);
-      
+
       return result;
    }
 
    /**
     * Try to find the resource locally
-    * 
+    *
     * @param name the resource name
     * @return the url if found
     */
@@ -570,7 +596,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
 
    /**
     * Try to find the resource locally
-    * 
+    *
     * @param name the resource name
     * @param trace whether trace is enabled
     * @return the URL if found
@@ -591,7 +617,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
             return url;
          }
       }
-      
+
       // Is this resource blacklisted?
       if (blackList != null && blackList.containsKey(name))
       {
@@ -599,7 +625,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
             log.trace(this + " resource is blacklisted " + name);
          return null;
       }
-      
+
       // Ask the policy for the resource
       URL result = AccessController.doPrivileged(new PrivilegedAction<URL>()
       {
@@ -622,17 +648,17 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
       // Cache what we found
       if (resourceCache != null && result != null)
          resourceCache.put(name, result);
-      
+
       // Blacklist when not found
       if (blackList != null && result == null)
          blackList.put(name, name);
-      
+
       return result;
    }
 
    /**
     * Try to find the resource locally
-    * 
+    *
     * @param name the resource name
     * @param urls the urls to add to
     * @throws IOException for any error
@@ -644,7 +670,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
 
    /**
     * Try to find the resources locally
-    * 
+    *
     * @param name the resource name
     * @param urls the urls to add to
     * @param trace whether trace is enabled
@@ -679,7 +705,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
          throw e2;
       }
    }
-   
+
    /**
     * Define the package for the class if not already done
     *
@@ -691,10 +717,10 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
       String packageName = ClassLoaderUtils.getClassPackageName(className);
       if (packageName.length() == 0)
          return;
-      
+
       // Ask the policy for the information
       PackageInformation pi = policy.getClassPackageInformation(className, packageName);
-      
+
       // Already defined?
       Package pkge = getPackage(packageName);
       if (pkge != null)
@@ -731,7 +757,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
 
    /**
     * Try to load the class from the domain
-    * 
+    *
     * @param name the class name
     * @param trace whether trace is enabled
     * @return the class if found in the parent
@@ -741,17 +767,17 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
    {
       // Because of the broken classloading in the Sun JDK we need to
       // serialize access to the classloader.
-      
+
       // Additionally, acquiring the lock on the policy for this classloader
       // ensures that we don't race with somebody undeploying the classloader
       // which could cause leaks
       acquireLockFairly(trace);
       try
       {
-         // Here we have synchronized with the policy 
+         // Here we have synchronized with the policy
          BaseClassLoaderPolicy basePolicy = policy;
          BaseClassLoaderDomain domain = basePolicy.getClassLoaderDomain();
-         
+
          if (trace)
             log.trace(this + " load from domain " + name + " domain=" + domain);
 
@@ -761,7 +787,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
          if (domain == null)
          {
             Class<?> result = loadClassLocally(name, trace);
-            
+
             // So this is almost certainly a classloader leak
             if (result == null)
                throw new IllegalStateException(this + " classLoader is not connected to a domain (probably undeployed?) for class " + name);
@@ -776,7 +802,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
       }
       finally
       {
-         unlock(trace);
+         unlock(trace, true);
       }
    }
 
@@ -785,7 +811,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
       BaseClassLoaderPolicy basePolicy = policy;
       return basePolicy.getClassLoaderUnchecked() != null;
    }
-   
+
    public Class<?> getCachedClass(String name)
    {
       // TODO look in global and/or local cache
@@ -850,17 +876,17 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
          policy.clearBlackList(name);
       }
    }
-   
+
    /**
     * A long version of the classloader
-    * 
+    *
     * @return the long string
     */
    public String toLongString()
    {
       if (cachedToString != null)
          return cachedToString;
-      
+
       StringBuilder builder = new StringBuilder();
       builder.append(getClass().getSimpleName());
       builder.append('@').append(Integer.toHexString(System.identityHashCode(this)));
@@ -870,7 +896,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
       cachedToString = builder.toString();
       return cachedToString;
    }
-   
+
    /**
     * Shutdown the classloader
     */
@@ -883,16 +909,16 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
          blackList.clear();
       cachedToString = null;
    }
-   
+
    /**
     * For subclasses to add things to the long string
-    * 
+    *
     * @param builder the builder
     */
    protected void toLongString(StringBuilder builder)
    {
    }
-   
+
    @Override
    public String toString()
    {
@@ -905,7 +931,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
 
    /**
     * Attempt to lock, but don't wait
-    * 
+    *
     * @return true when the lock was obtained
     */
    boolean attemptLock()
@@ -915,7 +941,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
 
    /**
     * Lock
-    * 
+    *
     * This method must be invoked with the monitor held
     */
    void lock()
@@ -925,15 +951,17 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
 
    /**
     * Unlock
+    *
+    * @param rescheduleTasks whether to reschedule tasks
     */
-   void unlock()
+   void unlock(boolean rescheduleTasks)
    {
-      unlock(log.isTraceEnabled());
+      unlock(log.isTraceEnabled(), rescheduleTasks);
    }
 
    /**
     * Attempt to get the lock but don't wait
-    * 
+    *
     * @param trace whether trace is enabled
     * @return true when obtained the lock
     */
@@ -952,6 +980,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
       }
       catch (InterruptedException ignored)
       {
+         interrupted = true;
       }
       finally
       {
@@ -972,12 +1001,12 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
 
       return result;
    }
-   
+
    /**
     * Acquire the lock on the classloader fairly<p>
     *
     * This must be invoked with the monitor held
-    * 
+    *
     * @param trace whether trace is enabled
     */
    private void acquireLockFairly(boolean trace)
@@ -989,7 +1018,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
       boolean interrupted = Thread.interrupted();
 
       int waits = 0;
-      
+
       try
       {
          while (true)
@@ -1015,6 +1044,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
             }
             catch (InterruptedException ignored)
             {
+               interrupted = true;
             }
          }
       }
@@ -1023,17 +1053,18 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
          if (interrupted)
             thread.interrupt();
       }
-      
+
       if (lock.getHoldCount() == 1)
          ClassLoaderManager.registerLoaderThread(this, thread);
    }
 
    /**
     * Unlock
-    * 
+    *
     * @param trace whether trace is enabled
+    * @param rescheduleTasks whether to reschedule tasks
     */
-   private void unlock(boolean trace)
+   private void unlock(boolean trace, boolean rescheduleTasks)
    {
       Thread thread = Thread.currentThread();
       if (trace)
@@ -1041,19 +1072,19 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
 
       synchronized (this)
       {
-         lock.unlock();      
-      
+         lock.unlock();
+
          if (lock.getHoldCount() == 0)
          {
-            ClassLoaderManager.unregisterLoaderThread(this, thread);
+            ClassLoaderManager.unregisterLoaderThread(this, thread, rescheduleTasks);
             notifyAll();
          }
       }
    }
-   
+
    /**
     * Get the classloader for a class
-    * 
+    *
     * @param clazz the class
     * @return the classloader or null if it doesn't have one
     */
@@ -1062,7 +1093,7 @@ public class BaseClassLoader extends SecureClassLoader implements BaseClassLoade
       SecurityManager sm = System.getSecurityManager();
       if (sm == null)
          return clazz.getClassLoader();
-      
+
       return AccessController.doPrivileged(new PrivilegedAction<ClassLoader>()
       {
          public ClassLoader run()
