@@ -68,7 +68,7 @@ public abstract class BaseClassLoaderDomain implements Loader
    private Map<String, List<ClassLoaderInformation>> classLoadersByPackageName = new ConcurrentHashMap<String, List<ClassLoaderInformation>>();
    
    /** The global class cache */
-   private Map<String, Loader> globalClassCache = new ConcurrentHashMap<String, Loader>();
+   private Map<String, ClassCacheItem> globalClassCache = new ConcurrentHashMap<String, ClassCacheItem>();
    
    /** The global class black list */
    private Map<String, String> globalClassBlackList = new ConcurrentHashMap<String, String>();
@@ -91,6 +91,9 @@ public abstract class BaseClassLoaderDomain implements Loader
       globalClassBlackList.clear();
       globalResourceCache.clear();
       globalResourceBlackList.clear();
+
+      for (ClassLoaderInformation info : classLoaders)
+         info.flushCaches();
    }
 
    public int getClassBlackListSize()
@@ -121,7 +124,7 @@ public abstract class BaseClassLoaderDomain implements Loader
    public Map<String, String> listClassCache()
    {
       Map<String, String> result = new HashMap<String, String>(globalClassCache.size());
-      for (Map.Entry<String, Loader> entry : globalClassCache.entrySet())
+      for (Map.Entry<String, ClassCacheItem> entry : globalClassCache.entrySet())
          result.put(entry.getKey(), entry.getValue().toString());
       return result;
    }
@@ -234,7 +237,12 @@ public abstract class BaseClassLoaderDomain implements Loader
    protected Class<?> loadClass(BaseClassLoader classLoader, String name, boolean allExports) throws ClassNotFoundException
    {
       boolean trace = log.isTraceEnabled();
+      
+      String path = ClassLoaderUtils.classNameToPath(name);
 
+      // JBCL-101 TODO need to rework the blacklist
+      //checkClassBlackList(classLoader, name, path, allExports, true);
+      
       boolean findInParent = (isUseLoadClassForParent() == false);
       
       // Should we directly load from the parent?
@@ -242,18 +250,23 @@ public abstract class BaseClassLoaderDomain implements Loader
       {
          Class<?> clazz = loadClassBefore(name);
          if (clazz != null)
+         {
+            globalClassCache.put(path, new ClassCacheItem(clazz));
             return clazz;
+         }
       }
       
-      String path = ClassLoaderUtils.classNameToPath(name);
-      
       Loader loader = findLoader(classLoader, path, allExports, findInParent);
       if (loader != null)
       {
          Thread thread = Thread.currentThread();
          ClassLoadingTask task = new ClassLoadingTask(name, classLoader, thread);
          ClassLoaderManager.scheduleTask(task, loader, false);
-         return ClassLoaderManager.process(thread, task);
+         Class<?> result = ClassLoaderManager.process(thread, task);
+         ClassCacheItem item = globalClassCache.get(path);
+         if (item != null)
+            item.clazz = result;
+         return result;
       }
       
       // Should we directly load from the parent?
@@ -261,7 +274,10 @@ public abstract class BaseClassLoaderDomain implements Loader
       {
          Class<?> clazz = loadClassAfter(name);
          if (clazz != null)
+         {
+            globalClassCache.put(path, new ClassCacheItem(clazz));
             return clazz;
+         }
       }
 
       // Finally see whether this is the JDK assuming it can load its classes from any classloader
@@ -278,6 +294,7 @@ public abstract class BaseClassLoaderDomain implements Loader
             {
                if (trace)
                   log.trace(this + " loaded from hack " + hack + " " + ClassLoaderUtils.classToString(result));
+               globalClassCache.put(path, new ClassCacheItem(result));
                return result;
             }
          }
@@ -624,13 +641,16 @@ public abstract class BaseClassLoaderDomain implements Loader
     */
    private Loader findLoaderInExports(BaseClassLoader classLoader, String name, boolean trace)
    {
-      Loader loader = globalClassCache.get(name);
-      if (loader != null)
+      ClassCacheItem item = globalClassCache.get(name);
+      if (item != null)
       {
-         if (trace)
-            log.trace(this + " found in global class cache " + name);
-
-         return loader;
+         Loader loader = item.loader;
+         if (loader != null)
+         {
+            if (trace)
+               log.trace(this + " found loader " + loader + " in global class cache " + name);
+            return loader;
+         }
       }
 
       if (globalClassBlackList.containsKey(name))
@@ -662,7 +682,7 @@ public abstract class BaseClassLoaderDomain implements Loader
             if (exported.getResource(name) != null)
             {
                if (canCache)
-                  globalClassCache.put(name, exported);
+                  globalClassCache.put(name, new ClassCacheItem(exported));
                return exported;
             }
          }
@@ -1415,6 +1435,83 @@ public abstract class BaseClassLoaderDomain implements Loader
       return result;
    }
 
+   /**
+    * Check the class cache
+    * 
+    * @param classLoader the reference classloader (possibly null)
+    * @param name the name of the class
+    * @param path the path of the class resource
+    * @param allExports whether to look at all exports
+    * @return the class if cached
+    */
+   Class<?> checkClassCache(BaseClassLoader classLoader, String name, String path, boolean allExports)
+   {
+      if (allExports)
+      {
+         ClassCacheItem item = globalClassCache.get(path);
+         if (item != null)
+         {
+            if (log.isTraceEnabled())
+               log.trace("Found " + name + " in global cache");
+            return item.clazz;
+         }
+      }
+      return null;
+   }
+
+   /**
+    * Check the class blacklist
+    * 
+    * @param classLoader the classloader (possibly null)
+    * @param name the name
+    * @param path the path of the class resource
+    * @param allExports whether to look at all exports
+    * @param failIfBlackListed <code>true</code> if a blacklisted class should
+    *                          result in ClassNotFoundException; <code>false</code>
+    *                          if a <code>null</code> return value is acceptable
+    * @throws ClassNotFoundException when the class is blacklisted
+    */
+   void checkClassBlackList(BaseClassLoader classLoader, String name, String path, boolean allExports, boolean failIfBlackListed) throws ClassNotFoundException
+   {
+      if (allExports)
+      {
+         if (failIfBlackListed && globalClassBlackList.containsKey(path))
+         {
+            if (log.isTraceEnabled())
+               log.trace("Found " + name + " in global blacklist");
+            throw new ClassNotFoundException(name + " not found - blacklisted");
+         }
+      }
+   }
+   
+   /**
+    * Check the cache and blacklist
+    * 
+    * @param classLoader the classloader (possibly null)
+    * @param name the name
+    * @param path the path of the class resource
+    * @param allExports whether to look at all exports
+    * @param failIfBlackListed <code>true</code> if a blacklisted class should
+    *                          result in ClassNotFoundException; <code>false</code>
+    *                          if a <code>null</code> return value is acceptable
+    * @return the class when found in the cache
+    * @throws ClassNotFoundException when the class is blacklisted and 
+    *                               <code>failIfBlackListed</code> is <code>true</code>
+    */
+   protected Class<?> checkClassCacheAndBlackList(BaseClassLoader classLoader, String name, String path, boolean allExports, boolean failIfBlackListed) throws ClassNotFoundException
+   {
+      if (path == null)
+         path = ClassLoaderUtils.classNameToPath(name);
+      
+      Class<?> result = checkClassCache(classLoader, name, path, allExports);
+      if (result != null)
+         return result;
+      
+      checkClassBlackList(classLoader, name, path, allExports, failIfBlackListed);
+      
+      return null;
+   }
+
    /**
     * Cleans the entry with the given name from the blackList
     *
@@ -1430,6 +1527,37 @@ public abstract class BaseClassLoaderDomain implements Loader
       {
          globalResourceBlackList.remove(name);
       }
+
+      // Need to clear the import caches as well
+      List<ClassLoaderInformation> infos = classLoaders;
+      for (ClassLoaderInformation info : infos)
+         info.clearBlackList(name);
    }
    
+   /**
+    * ClassCacheItem.
+    */
+   class ClassCacheItem
+   {
+      Loader loader;
+      Class<?> clazz;
+      
+      public ClassCacheItem(Loader loader)
+      {
+         this.loader = loader;
+      }
+      
+      public ClassCacheItem(Class<?> clazz)
+      {
+         this.clazz = clazz;
+      }
+
+      @Override
+      public String toString()
+      {
+         if (loader != null)
+            return loader.toString();
+         return null;
+      }
+   }
 }
